/*
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/*
 *    Author Quan Sun (quan.sun.nz@gmail.com)
 *    Copyright (C) 2013 University of Waikato, Hamilton, New Zealand
 *
 */
package artforests.core;

import weka.core.AbstractInstance;
import weka.core.Instance;
import weka.core.RevisionUtils;
import weka.core.Utils;

public class MultiTargetInstance extends AbstractInstance {

    private double[] m_RankVector;
    
    private String m_InstanceLabel;

    public String getInstanceLabel() {
        return m_InstanceLabel;
    }

    public void setInstanceLabel(String l) {
        m_InstanceLabel = l;
    }

    public double[] getRanking() {
        double[] r = new double[m_RankVector.length];
        System.arraycopy(m_RankVector, 0, r, 0, r.length);
        return r;
    }

    public void setRanking(double[] ranking) {
        m_RankVector = new double[ranking.length];
        System.arraycopy(ranking, 0, m_RankVector, 0, ranking.length);
    }

    public MultiTargetInstance(/*@non_null@*/Instance instance) {

        if (instance instanceof MultiTargetInstance) {
            m_AttValues = ((MultiTargetInstance) instance).m_AttValues;
        } else {
            m_AttValues = instance.toDoubleArray();
        }
        m_Weight = instance.weight();
        m_Dataset = null;
    }

    public MultiTargetInstance(double weight, /*@non_null@*/ double[] attValues) {

        m_AttValues = attValues;
        m_Weight = weight;
        m_Dataset = null;
    }

    public MultiTargetInstance(int numAttributes) {

        m_AttValues = new double[numAttributes];
        for (int i = 0; i < m_AttValues.length; i++) {
            m_AttValues[i] = Utils.missingValue();
        }
        m_Weight = 1;
        m_Dataset = null;
    }

    public /*@pure@*/ Object copy() {

        MultiTargetInstance result = new MultiTargetInstance(this);
        result.m_Dataset = m_Dataset;
        //result.setID(m_ID);
        result.setRanking(m_RankVector);        
        return result;
    }

    public /*@pure@*/ int index(int position) {

        return position;
    }

    public Instance mergeInstance(Instance inst) {

        int m = 0;
        double[] newVals = new double[numAttributes() + inst.numAttributes()];
        for (int j = 0; j < numAttributes(); j++, m++) {
            newVals[m] = value(j);
        }
        for (int j = 0; j < inst.numAttributes(); j++, m++) {
            newVals[m] = inst.value(j);
        }
        return new MultiTargetInstance(1.0, newVals);
    }

    public /*@pure@*/ int numAttributes() {

        return m_AttValues.length;
    }

    public /*@pure@*/ int numValues() {

        return m_AttValues.length;
    }

    public void replaceMissingValues(double[] array) {

        if ((array == null)
                || (array.length != m_AttValues.length)) {
            throw new IllegalArgumentException("Unequal number of attributes!");
        }
        freshAttributeVector();
        for (int i = 0; i < m_AttValues.length; i++) {
            if (isMissing(i)) {
                m_AttValues[i] = array[i];
            }
        }
    }

    public void setValue(int attIndex, double value) {

        freshAttributeVector();
        m_AttValues[attIndex] = value;
    }

    public void setValueSparse(int indexOfIndex, double value) {

        freshAttributeVector();
        m_AttValues[indexOfIndex] = value;
    }

    public double[] toDoubleArray() {

        double[] newValues = new double[m_AttValues.length];
        System.arraycopy(m_AttValues, 0, newValues, 0,
                m_AttValues.length);
        return newValues;
    }

    public String toStringNoWeight() {
        StringBuffer text = new StringBuffer();

        for (int i = 0; i < m_AttValues.length; i++) {
            if (i > 0) {
                text.append(",");
            }
            text.append(toString(i));
        }

        return text.toString();
    }

    public /*@pure@*/ double value(int attIndex) {

        return m_AttValues[attIndex];
    }

    protected void forceDeleteAttributeAt(int position) {

        double[] newValues = new double[m_AttValues.length - 1];

        System.arraycopy(m_AttValues, 0, newValues, 0, position);
        if (position < m_AttValues.length - 1) {
            System.arraycopy(m_AttValues, position + 1,
                    newValues, position,
                    m_AttValues.length - (position + 1));
        }
        m_AttValues = newValues;
    }

    protected void forceInsertAttributeAt(int position) {

        double[] newValues = new double[m_AttValues.length + 1];

        System.arraycopy(m_AttValues, 0, newValues, 0, position);
        newValues[position] = Utils.missingValue();
        System.arraycopy(m_AttValues, position, newValues,
                position + 1, m_AttValues.length - position);
        m_AttValues = newValues;
    }

    private void freshAttributeVector() {

        m_AttValues = toDoubleArray();
    }

    public String getRevision() {
        return RevisionUtils.extract("$Revision: 1234 $");
    }

}
